package gregapi.block.multitileentity;

import static gregapi.data.CS.W;
import gregapi.GT_API;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_OnRegistration;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_OnRegistrationClient;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.ItemStackContainer;
import gregapi.code.ItemStackMap;
import gregapi.lang.LanguageHandler;
import gregapi.util.UT;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.minecraft.block.Block;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;

/**
 * @author Gregorius Techneticies
 * 
 * Create yourself a new MultiTileEntity Registry in order to have your very own 32767 Sub IDs.
 * You can do with those IDs whatever you want since this automatically creates your personal Item and DOESN'T use any Items or Blocks of GregTech itself.
 * 
 * ================================================================================================================================================
 * The way this whole System works is very simple. The setTileEntity call can set the TileEntity of your choice at every Location you want.
 * If now the BlockContainer doesn't return a TileEntity, but instead the ItemBlock manually sets the TileEntity, you can have every single
 * TileEntity being placed at every Block you want. If that Block then is compatible with your TileEntity (via Interfaces and such) it can
 * easily make use of the TileEntity no matter which one it is.
 * 
 * "But what is with the Loading of those TileEntities? Don't they get deleted on startup?" You think? No they don't get deleted. Minecraft
 * can load every TileEntity just by a Name->Class Map (you know about that when you have ever created a TileEntity yourself), and the remaining
 * Stats can be saved inside the NBT of the TileEntity.
 * 
 * In the end I have a dynamic collection of Blocks to get the vanilla Materials and Sound Effects right, a Registry of TileEntities to be
 * attached to those Blocks via additional custom ItemBlocks to enable everything, and an automatic Network Handler.
 * 
 * The only thing needed to be done manually is something that transmits the Data from the Server to the Client to set the proper TileEntity there.
 * 
 * In order to do that, just send one of the 5 Packets (PacketSyncDataByteAndIDs, PacketSyncDataShortAndIDs, PacketSyncDataIntegerAndIDs, 
 * PacketSyncDataLongAndIDs or PacketSyncDataByteArrayAndIDs) for transmitting the ID to the Client with aID1 = getMultiTileEntityRegistryID() and
 * aID2 = getMultiTileEntityID(). The Byte/Short/Integer/Long/ByteArray can be used for transmitting other Data, such as a Facing to the Client.
 * ================================================================================================================================================
 */
public class MultiTileEntityRegistry {
	private static final ItemStackMap<ItemStackContainer, MultiTileEntityRegistry> REGISTRIES = new ItemStackMap<ItemStackContainer, MultiTileEntityRegistry>();
	
	public Map<Short, MultiTileEntityClassContainer> mRegistry = new HashMap<Short, MultiTileEntityClassContainer>();
	public List<MultiTileEntityClassContainer> mRegistrations = new ArrayListNoNulls<MultiTileEntityClassContainer>();
	
	public final String mNameInternal;
	public final MultiTileEntityBlockInternal mBlock;
	
	/** 
	 * @param aNameInternal the internal Name of the Item. DO NOT START YOUR UNLOCALISED NAME WITH "gt."!!!
	 */
	public MultiTileEntityRegistry(String aNameInternal) {
		if (!GT_API.sAPIPhaseStartedPreLoad || GT_API.sAPIPhaseStartedLoad) throw new IllegalStateException("The MultiTileEntity Registry must be initialised at the Preload Phase, because it relies on an ItemBlock being created.");
		mNameInternal = aNameInternal;
        mBlock = new MultiTileEntityBlockInternal(this);
        REGISTRIES.put(new ItemStackContainer(mBlock, 1, W), this);
	}
	
	public static MultiTileEntityRegistry getRegistry(int aRegistryID) {
		return REGISTRIES.get(new ItemStackContainer(Item.getItemById(aRegistryID), 1, W));
	}
	
	/**
	 * Adds a new MultiTileEntity.
	 */
	public ItemStack add(String aLocalised, MultiTileEntityClassContainer aClassContainer) {
		if (aClassContainer == null || aClassContainer.mID == W || aClassContainer.mID < 0 || mRegistry.containsKey(aClassContainer.mID)) return null;
		LanguageHandler.add(mNameInternal+"."+aClassContainer.mID+".name", aLocalised);
		mRegistry.put(aClassContainer.mID, aClassContainer);
		mRegistrations.add(aClassContainer);
		if (aClassContainer.mCanonicalTileEntity instanceof IMTE_OnRegistration) {
			((IMTE_OnRegistration)aClassContainer.mCanonicalTileEntity).onRegistration(this);
			if (GT_API.api_proxy.isClientSide() && aClassContainer.mCanonicalTileEntity instanceof IMTE_OnRegistrationClient) ((IMTE_OnRegistrationClient)aClassContainer.mCanonicalTileEntity).onRegistrationClient(this);
		}
		return getItem(aClassContainer.mID);
	}
	
	public String getLocal(int aID) {
		return LanguageHandler.translate(mNameInternal+"."+aID+".name");
	}
	
	public ItemStack getItem(int aID) {
		return new ItemStack(mBlock, 1, aID);
	}
	
	public ItemStack getItem(int aID, NBTTagCompound aNBT) {
		ItemStack rStack = new ItemStack(mBlock, 1, aID);
		rStack.setTagCompound(aNBT);
		return rStack;
	}
	
	public ItemStack getItem(int aID, long aAmount) {
		return new ItemStack(mBlock, (int)aAmount, aID);
	}
	
	public ItemStack getItem(int aID, long aAmount, NBTTagCompound aNBT) {
		ItemStack rStack = new ItemStack(mBlock, (int)aAmount, aID);
		rStack.setTagCompound(aNBT);
		return rStack;
	}
	
	public TileEntity getNewTileEntity(int aID) {
		MultiTileEntityContainer tContainer = getNewTileEntityContainer(aID, null);
		return tContainer == null ? null : tContainer.mTileEntity;
	}
	
	public TileEntity getNewTileEntity(ItemStack aStack) {
		MultiTileEntityContainer tContainer = getNewTileEntityContainer(aStack);
		return tContainer == null ? null : tContainer.mTileEntity;
	}
	
	public MultiTileEntityContainer getNewTileEntityContainer(int aID, NBTTagCompound aNBT) {
		MultiTileEntityClassContainer tClass = mRegistry.get((short)aID);
		if (tClass == null) return null;
		MultiTileEntityContainer rContainer = new MultiTileEntityContainer((TileEntity)UT.Reflection.callConstructor(tClass.mClass, -1, null, true), tClass.mBlock, tClass.mBlockMetaData);
		if (rContainer.mTileEntity == null || rContainer.mBlock == null) return null;
		((IMultiTileEntity)rContainer.mTileEntity).initFromNBT(aNBT == null ? tClass.mParameters : UT.NBT.fuseNBTs(tClass.mParameters, aNBT), (short)aID, (short)Block.getIdFromBlock(mBlock));
		return rContainer;
	}
	
	public MultiTileEntityContainer getNewTileEntityContainer(ItemStack aStack) {
		return getNewTileEntityContainer(aStack.getItemDamage(), aStack.getTagCompound());
	}
	
	public MultiTileEntityClassContainer getClassContainer(int aID) {
		return mRegistry.get((short)aID);
	}
	public MultiTileEntityClassContainer getClassContainer(ItemStack aStack) {
		return mRegistry.get((short)aStack.getItemDamage());
	}
}